(* ICFP Programming Contest 2001
 * Copyright (C) 2001 Jean-Christophe Filliatre *)

(*s Lexer for SML/NG documents. A little trick here: many spaces are 
    replaced by a single one outside of any \verb!TT! tag (a counter
    [tt_level] counting the \verb!TT! depth is used for that purpose). 
    Of course, some other spaces may be compressed if, for instance,
    a \verb!PL! tag is used inside a \verb!TT! tag; this will be treated
    later. But this trick gives us an ``initial'' document already a
    little smaller; note that this document will be printed on timeout,
    not the exact original one. *)

{
  open Lexing
  open Parser
  open Sml_ng

  (*s Statically allocated tags (see remark in module [Parser]). *)

  let spc = WSP SPC
  let tab = WSP TAB
  let lf = WSP LF
  let cr = WSP CR
  let size_0 = Size 0
  let size_1 = Size 1
  let size_2 = Size 2
  let size_3 = Size 3
  let size_4 = Size 4
  let size_5 = Size 5
  let size_6 = Size 6
  let size_7 = Size 7
  let size_8 = Size 8
  let size_9 = Size 9
  let red = Color Red
  let green = Color Green
  let blue = Color Blue
  let cyan = Color Cyan
  let magenta = Color Magenta
  let yellow = Color Yellow
  let black = Color Black
  let white = Color White

  (*s The \verb!TT! depth. *)

  let tt_level = ref 0

}

(*s Shortcuts for characters. *)

let spc = ' '
let tab = '\t'
let lf = '\n'
let cr = '\r'
let space = spc | tab | lf | cr
let char = ['\033'-'\059' '\061' '\063'-'\126']

(*s Rules. *)

rule token = parse
  | spc { if !tt_level == 0 then space lexbuf else spc }
  | tab { if !tt_level == 0 then space lexbuf else tab }
  | lf  { if !tt_level == 0 then space lexbuf else lf }
  | cr  { if !tt_level == 0 then space lexbuf else cr }
  (* Opening tags *)
  | "<B>" { OPEN B }
  | "<EM>" { OPEN EM }
  | "<I>" { OPEN I }
  | "<PL>" { OPEN PL }
  | "<S>" { OPEN S }
  | "<TT>" { incr tt_level; OPEN TT }
  | "<U>" { OPEN U }
  | "<0>" { OPEN size_0 }
  | "<1>" { OPEN size_1 }
  | "<2>" { OPEN size_2 }
  | "<3>" { OPEN size_3 }
  | "<4>" { OPEN size_4 }
  | "<5>" { OPEN size_5 }
  | "<6>" { OPEN size_6 }
  | "<7>" { OPEN size_7 }
  | "<8>" { OPEN size_8 }
  | "<9>" { OPEN size_9 }
  | "<r>" { OPEN red }
  | "<g>" { OPEN green }
  | "<b>" { OPEN blue }
  | "<c>" { OPEN cyan }
  | "<m>" { OPEN magenta }
  | "<y>" { OPEN yellow }
  | "<k>" { OPEN black }
  | "<w>" { OPEN white }
  (* Closing tags *)
  | "</B>" { CLOSE B }
  | "</EM>" { CLOSE EM }
  | "</I>" { CLOSE I }
  | "</PL>" { CLOSE PL }
  | "</S>" { CLOSE S }
  | "</TT>" { decr tt_level; CLOSE TT }
  | "</U>" { CLOSE U }
  | "</0>" { CLOSE size_0 }
  | "</1>" { CLOSE size_1 }
  | "</2>" { CLOSE size_2 }
  | "</3>" { CLOSE size_3 }
  | "</4>" { CLOSE size_4 }
  | "</5>" { CLOSE size_5 }
  | "</6>" { CLOSE size_6 }
  | "</7>" { CLOSE size_7 }
  | "</8>" { CLOSE size_8 }
  | "</9>" { CLOSE size_9 }
  | "</r>" { CLOSE red }
  | "</g>" { CLOSE green }
  | "</b>" { CLOSE blue }
  | "</c>" { CLOSE cyan }
  | "</m>" { CLOSE magenta }
  | "</y>" { CLOSE yellow }
  | "</k>" { CLOSE black }
  | "</w>" { CLOSE white }
  (* Other chars *)
  | char+ { TEXT (lexeme lexbuf) }
  | eof { EOF }
  | _ { raise Parsing.Parse_error }

and space = parse
  | space+ { spc }
  | eof    { lexbuf.lex_curr_pos <- lexbuf.lex_start_pos; spc }
  | _      { lexbuf.lex_curr_pos <- lexbuf.lex_start_pos; spc }

