(* ICFP Programming Contest 2001
 * Copyright (C) 2001 Jean-Christophe Filliatre *)

(*s Output of a document. Used to output the original document only,
    if timeout is reached or optimization didn't give something better. 
    Notice that a color shouldn't be [Any] and that size shouldn't be
    $-1$ (an assertion checks it). *)

open Sml_ng

let print_color = function
  | Red -> print_char 'r'
  | Green -> print_char 'g'
  | Blue -> print_char 'b'
  | Cyan -> print_char 'c'
  | Magenta -> print_char 'm' 
  | Yellow -> print_char 'y' 
  | Black -> print_char 'k'
  | White -> print_char 'w'
  | Any -> assert false

let open_tag = function
  | B -> print_string "<B>"
  | EM -> print_string "<EM>"
  | I -> print_string "<I>"
  | PL -> print_string "<PL>"
  | S -> print_string "<S>"
  | TT -> print_string "<TT>"
  | U -> print_string "<U>"
  | Size n -> assert (n != -1); print_char '<'; print_int n; print_char '>'
  | Color c -> print_char '<'; print_color c; print_char '>'

let close_tag = function
  | B -> print_string "</B>"
  | EM -> print_string "</EM>"
  | I -> print_string "</I>"
  | PL -> print_string "</PL>"
  | S -> print_string "</S>"
  | TT -> print_string "</TT>"
  | U -> print_string "</U>"
  | Size n -> assert (n != -1); print_string "</"; print_int n; print_char '>'
  | Color c -> print_string "</"; print_color c; print_char '>'

let rec output_item = function
  | Text s     -> print_string s
  | Wspace SPC -> print_char ' '
  | Wspace TAB -> print_char '\t'
  | Wspace CR  -> print_char '\013'
  | Wspace LF  -> print_char '\010'
  | Tag (t, d) -> open_tag t; List.iter output_item d; close_tag t

let output_document = List.iter output_item

(*s Size of a document (when output, in bytes). *)

let tag_size = function
  | EM | PL | TT -> 9
  | _ -> 7

let rec size_item n = function
  | Text s -> n + String.length s
  | Wspace _ -> succ n
  | Tag (t, d) -> n + tag_size t + size d

and size d = List.fold_left size_item 0 d

(*s Dumping of interpretations. *)

open Printf

let color_name = function
  | Red -> 'r'
  | Green -> 'g'
  | Blue -> 'b'
  | Cyan -> 'c'
  | Magenta -> 'm' 
  | Yellow -> 'y' 
  | Black -> 'k'
  | White -> 'w'
  | Any -> '-'

let dump_char ch (c,ctx) =
  fprintf ch 
    "'%s' (code %d) B=%b EM=%b I=%b S=%b TT=%b u=%d size=%d color=%c\n"
    (Char.escaped c) (Char.code c) 
    ctx.b ctx.em ctx.i ctx.s ctx.tt ctx.u ctx.size (color_name ctx.color)

let dump ch s = Stream.iter (dump_char ch) s; flush ch
